/***** Start of precommondefs.h *****/

/* This is extracted from pyconfig.h from CPython.  It sets the macros
   that affect the features we get from system include files.
   It must not #include anything. */

#ifndef __PYPY_PRECOMMONDEFS_H
#define __PYPY_PRECOMMONDEFS_H


/* Define on Darwin to activate all library features */
#define _DARWIN_C_SOURCE 1
/* This must be set to 64 on some systems to enable large file support. */
#define _FILE_OFFSET_BITS 64
/* Define on Linux to activate all library features */
#define _GNU_SOURCE 1
/* This must be defined on some systems to enable large file support. */
#define _LARGEFILE_SOURCE 1
/* Define on NetBSD to activate all library features */
#define _NETBSD_SOURCE 1
/* Define to activate features from IEEE Stds 1003.1-2001 */
#ifndef _POSIX_C_SOURCE
#  define _POSIX_C_SOURCE 200112L
#endif
/* Define on FreeBSD to activate all library features */
#define __BSD_VISIBLE 1
#define __XSI_VISIBLE 700
/* Windows: winsock/winsock2 mess */
#define WIN32_LEAN_AND_MEAN
#ifdef _WIN64
   typedef          __int64 Signed;
   typedef unsigned __int64 Unsigned;
#  define SIGNED_MIN LLONG_MIN
#else
   typedef          long Signed;
   typedef unsigned long Unsigned;
#  define SIGNED_MIN LONG_MIN
#endif

#if !defined(RPY_ASSERT) && !defined(RPY_LL_ASSERT)
#  define NDEBUG
#endif


/* All functions and global variables declared anywhere should use
   one of the following attributes:

   RPY_EXPORTED:  the symbol is exported out of libpypy-c.so.

   RPY_EXTERN:    the symbol is not exported out of libpypy-c.so, but
                  otherwise works like 'extern' by being available to
                  other C sources.

   static:        as usual, this means the symbol is local to this C file.

   Don't use _RPY_HIDDEN directly.  For tests involving building a custom
   .so, translator/tool/cbuild.py overrides RPY_EXTERN so that it becomes
   equal to RPY_EXPORTED.

   Any function or global variable declared with no attribute at all is
   a bug; please report or fix it.
*/
#ifdef __GNUC__
#  define RPY_EXPORTED extern __attribute__((visibility("default")))
#  define _RPY_HIDDEN  __attribute__((visibility("hidden")))
#else
#  define RPY_EXPORTED extern __declspec(dllexport)
#  define _RPY_HIDDEN  /* nothing */
#endif
#ifndef RPY_EXTERN
#  define RPY_EXTERN   extern _RPY_HIDDEN
#endif


#endif /* __PYPY_PRECOMMONDEFS_H */

/***** End of precommondefs.h *****/

#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <sys/select.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/timeb.h>

#include <stdio.h>
#include <stddef.h>   /* for offsetof() */

void dump(char* key, int value) {
    printf("%s: %d\n", key, value);
}


void dump_section_CLOCK_PROCESS_CPUTIME_ID(void) {
#ifdef CLOCK_PROCESS_CPUTIME_ID
	dump("defined", 1);
	if ((CLOCK_PROCESS_CPUTIME_ID) <= 0) {
	    long long x = (long long)(CLOCK_PROCESS_CPUTIME_ID);
	    printf("value: %lld\n", x);
	} else {
	    unsigned long long x = (unsigned long long)(CLOCK_PROCESS_CPUTIME_ID);
	    printf("value: %llu\n", x);
	}
#else
	dump("defined", 0);
#endif
}

void dump_section_EINTR(void) {
#ifdef EINTR
	dump("defined", 1);
	if ((EINTR) <= 0) {
	    long long x = (long long)(EINTR);
	    printf("value: %lld\n", x);
	} else {
	    unsigned long long x = (unsigned long long)(EINTR);
	    printf("value: %llu\n", x);
	}
#else
	dump("defined", 0);
#endif
}

void dump_section_GETTIMEOFDAY_NO_TZ(void) {
#ifdef GETTIMEOFDAY_NO_TZ
	dump("defined", 1);
#else
	dump("defined", 0);
#endif
}

void dump_section_RUSAGE(void) {
	typedef struct rusage platcheck_t;
	typedef struct {
	    char c;
	    platcheck_t s;
	} platcheck2_t;

	platcheck_t s;
	dump("align", offsetof(platcheck2_t, s));
	dump("size",  sizeof(platcheck_t));
	dump("fldofs ru_utime", offsetof(platcheck_t, ru_utime));
	dump("fldsize ru_utime",   sizeof(s.ru_utime));
	dump("fldofs ru_stime", offsetof(platcheck_t, ru_stime));
	dump("fldsize ru_stime",   sizeof(s.ru_stime));
}

void dump_section_RUSAGE_SELF(void) {
#ifdef RUSAGE_SELF
	dump("defined", 1);
	if ((RUSAGE_SELF) <= 0) {
	    long long x = (long long)(RUSAGE_SELF);
	    printf("value: %lld\n", x);
	} else {
	    unsigned long long x = (unsigned long long)(RUSAGE_SELF);
	    printf("value: %llu\n", x);
	}
#else
	dump("defined", 0);
#endif
}

void dump_section_TIMEVAL(void) {
	typedef struct timeval platcheck_t;
	typedef struct {
	    char c;
	    platcheck_t s;
	} platcheck2_t;

	platcheck_t s;
	dump("align", offsetof(platcheck2_t, s));
	dump("size",  sizeof(platcheck_t));
	dump("fldofs tv_sec", offsetof(platcheck_t, tv_sec));
	dump("fldsize tv_sec",   sizeof(s.tv_sec));
	s.tv_sec = 0; s.tv_sec = ~s.tv_sec;
	dump("fldunsigned tv_sec", s.tv_sec > 0);
	dump("fldofs tv_usec", offsetof(platcheck_t, tv_usec));
	dump("fldsize tv_usec",   sizeof(s.tv_usec));
	s.tv_usec = 0; s.tv_usec = ~s.tv_usec;
	dump("fldunsigned tv_usec", s.tv_usec > 0);
}

int main(int argc, char *argv[]) {
	printf("-+- CLOCK_PROCESS_CPUTIME_ID\n");
	dump_section_CLOCK_PROCESS_CPUTIME_ID();
	printf("---\n");
	printf("-+- EINTR\n");
	dump_section_EINTR();
	printf("---\n");
	printf("-+- GETTIMEOFDAY_NO_TZ\n");
	dump_section_GETTIMEOFDAY_NO_TZ();
	printf("---\n");
	printf("-+- RUSAGE\n");
	dump_section_RUSAGE();
	printf("---\n");
	printf("-+- RUSAGE_SELF\n");
	dump_section_RUSAGE_SELF();
	printf("---\n");
	printf("-+- TIMEVAL\n");
	dump_section_TIMEVAL();
	printf("---\n");
	return 0;
}
